from datetime import date, datetime
import numpy as np
import os
import pandas as pd


DATA_FOLDER = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'preparing master dataset')


# Step 1: Create a dataframe with all dates for which we create the master dataset
start_date = '2009-01-01'
end_date = datetime.now().date()
print('Adding dates')
# the resulting master_dataset_df will have 1 column with beginning of the month dates between start_date and end_date
master_dataset_df = pd.DataFrame()
master_dataset_df['date'] = pd.date_range(start=start_date, end=end_date, freq='MS')
master_dataset_df['year'] = master_dataset_df['date'].dt.year
master_dataset_df['month'] = master_dataset_df['date'].dt.month
master_dataset_df['date'] = master_dataset_df['date'].dt.date
master_dataset_df = master_dataset_df[master_dataset_df['date']<=date(2023, 7, 1)]

# Step 2: Add location data (governorate, district, ...)
print('Adding locations')
yemen_pcodes_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'yemen_pcodes.csv'))
master_dataset_df = master_dataset_df.merge(yemen_pcodes_df, how='cross')
master_dataset_df = master_dataset_df[~master_dataset_df['adm2_pcode'].isin(['YE1319', 'YE1324'])]
number_of_rows = master_dataset_df.shape[0]
number_of_columns = master_dataset_df.shape[1]

# Step 3: Add area of control adm2
print('Adding area of control')
yemen_area_of_control_adm2_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'yemen_areas_of_control_adm2.csv'))[['adm2_pcode', 'irg_or_dfa']]
master_dataset_df = master_dataset_df.merge(yemen_area_of_control_adm2_df, on=['adm2_pcode'], how='left')
assert number_of_rows == master_dataset_df.shape[0]

# Step 4: Add population
# https://data.humdata.org/dataset/worldpop-population-counts-for-yemen
# data from website above, where we took per year the geotiff file
if True:
    print('Adding population')
    population_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'yemen_adm2_population.csv'))[['year', 'month', 'adm2_pcode', 'estimated_population']]
    master_dataset_df = master_dataset_df.merge(population_df, on=['year', 'month', 'adm2_pcode'], how='left')
    # fill in all months with latest known population value (since population data is only once per year on 1 Jan)
    master_dataset_df['estimated_population'] = master_dataset_df.groupby(['adm2_pcode'])['estimated_population'].ffill()
    assert number_of_rows == master_dataset_df.shape[0]

# Step 5: Add FEWS NET IPC
# https://datacatalog.worldbank.org/search/dataset/0064614
if True:
    print('Adding fews')
    fews_df = pd.read_csv(os.path.join(DATA_FOLDER, 'FEWS February 2024 Update TrueBoundaries_04-19-24.csv'))[['country', 'admin_code', 'year', 'month', 'fews_ipc', 'fews_ha']]
    fews_df = fews_df[fews_df['country']=='Yemen']
    world_bank_fews_pcodes_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'world_bank_fews_pcodes.csv'))
    fews_df = fews_df.merge(world_bank_fews_pcodes_df, left_on='admin_code', right_on='pcode_fews', how='left')
    fews_df = fews_df[~fews_df['fews_ipc'].isna()]
    fews_df['fews_ha'] = fews_df['fews_ha'].fillna(0)

    fews_df['date'] = pd.to_datetime(dict(year=fews_df.year, month=fews_df.month, day=1)).dt.strftime('%Y-%m-%d')
    fews_df = fews_df[fews_df['date']<'2023-06-01']  # used for modeling

    fews_df = fews_df[['year', 'month', 'fews_ipc', 'fews_ha', 'adm2_pcode', 'perc_adm2_population']]
    fews_df['fews_ipc_weighted'] = fews_df['fews_ipc'] * fews_df['perc_adm2_population']
    df2 = fews_df.groupby(['year', 'month', 'adm2_pcode'])['fews_ha'].max().reset_index()
    df3 = fews_df.groupby(['year', 'month', 'adm2_pcode'])['fews_ipc_weighted'].sum().reset_index()
    fews_df = df2.merge(df3, on=['year', 'month', 'adm2_pcode'], how='outer')
    fews_df['fews_ipc_weighted'] = fews_df['fews_ipc_weighted'].round(2)
    master_dataset_df = master_dataset_df.merge(fews_df, on=['year', 'month', 'adm2_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 6: Add IPC
# https://data.humdata.org/dataset/b70c2734-2339-4a4d-a69d-fa2bd3225156
# note, this dataset is not used in the threshold modeling
if True:
    print('Adding ipc')
    ipc_df = pd.read_csv(os.path.join(DATA_FOLDER, 'ipc_district_2018-2023.csv'))

    ipc_df = ipc_df.rename(columns={
        'Date': 'date',
        'Pcode': 'adm2_pcode',
        'Phase 1': 'ipc_phase_1',
        'Phase 1 %': 'ipc_phase_1_perc',
        'Phase 2': 'ipc_phase_2',
        'Phase 2 %': 'ipc_phase_2_perc',
        'Phase 3': 'ipc_phase_3',
        'Phase 3 %': 'ipc_phase_3_perc',
        'Phase 4': 'ipc_phase_4',
        'Phase 4 %': 'ipc_phase_4_perc',
        'Phase 5': 'ipc_phase_5',
        'Phase 5 %': 'ipc_phase_5_perc',
        'Phase P3+': 'ipc_phase_3plus',
        'Phase P3+ %': 'ipc_phase_3plus_perc',
    }
    )[['date', 'adm2_pcode', 'ipc_phase_1', 'ipc_phase_1_perc', 'ipc_phase_2', 'ipc_phase_2_perc', 'ipc_phase_3', 'ipc_phase_3_perc', 'ipc_phase_4', 'ipc_phase_4_perc', 'ipc_phase_5', 'ipc_phase_5_perc', 'ipc_phase_3plus', 'ipc_phase_3plus_perc']]
    ipc_df['date'] = pd.to_datetime(ipc_df['date'], format='%m/%d/%Y')
    ipc_df['year'] = ipc_df['date'].dt.year
    ipc_df['month'] = ipc_df['date'].dt.month
    ipc_df = ipc_df.drop(columns=['date'], axis=1)
    ipc_df['ipc_phase'] = np.select(
        [ipc_df['ipc_phase_5_perc'] >= 0.2,
        ipc_df['ipc_phase_4_perc'] >= 0.2,
        ipc_df['ipc_phase_3_perc'] >= 0.2,
        ipc_df['ipc_phase_2_perc'] >= 0.2,
        ipc_df['ipc_phase_1_perc'] >= 0.2,
        True], 
        [5, 4, 3, 2, 1, 1])

    ipc_df = ipc_df.drop_duplicates(subset=['year', 'month', 'adm2_pcode'], keep='last')
    master_dataset_df = master_dataset_df.merge(ipc_df, on=['year', 'month', 'adm2_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 7: Add Telegram XR
# https://yemen.yeti.acaps.org/api/telegram_exchange_rates/?format=csv+%28non-paginated%29
if True:
    print('Adding telegram xr')
    telegram_df = pd.read_csv(os.path.join(DATA_FOLDER, 'api_telegram_exchange_rates.csv'))
    telegram_df['irg_or_dfa'] = np.select([telegram_df['location']=='aden', telegram_df['location']=='sanaa'], ['IRG', 'DFA'])
    telegram_df = telegram_df[telegram_df['currency']=='us_dollar']
    telegram_df['datetime'] = pd.to_datetime(telegram_df['datetime'], format='%Y-%m-%dT%H:%M:%SZ')
    telegram_df['year'] = telegram_df['datetime'].dt.year
    telegram_df['month'] = telegram_df['datetime'].dt.month
    telegram_df = telegram_df.rename(columns={'sell': 'xr'})
    telegram_df = telegram_df[['year', 'month', 'irg_or_dfa', 'xr']]
    telegram_df['datetime'] = pd.to_datetime(dict(year=telegram_df.year, month=telegram_df.month, day=1))
    telegram_df = telegram_df.groupby(['datetime', 'year', 'month', 'irg_or_dfa'])['xr'].mean().reset_index()
    df_telegram_before_2020_dfa = telegram_df[telegram_df['datetime']<'2019-12-01']
    df_telegram_after_2020_both = telegram_df[telegram_df['datetime']>='2019-12-01']
    df_telegram_before_2020_irg = df_telegram_before_2020_dfa.copy()
    df_telegram_before_2020_irg['irg_or_dfa'] = 'IRG'
    telegram_df = pd.concat([df_telegram_before_2020_dfa, df_telegram_before_2020_irg, df_telegram_after_2020_both]).sort_values(by=['year', 'month', 'irg_or_dfa'])
    telegram_df = telegram_df.drop(columns=['datetime'], axis=1)
    master_dataset_df = master_dataset_df.merge(telegram_df, on=['year', 'month', 'irg_or_dfa'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 8: Add ACLED conflict
# https://api.acleddata.com
# note, ACLED changes some entries backwards, only current version is available for download. The csv that we load was the current ACLED dataset at the time of modeling (August 2023)
if True:
    print('Adding acled conflict')
    acled_df = pd.read_csv(os.path.join(DATA_FOLDER, 'original_acled_conflict.csv'))
    acled_df = acled_df[acled_df['event_type'].isin(['Battles', 'Explosions/Remote violence', 'Violence against civilians'])]
    acled_df = acled_df[['event_date', 'admin2', 'fatalities']].rename(columns={'admin2': 'adm2_name_acled'})
    acled_pcodes_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'acled_pcodes.csv'))
    acled_df = acled_df.merge(acled_pcodes_df, on=['adm2_name_acled'], how='left').drop(columns=['adm2_name_acled'], axis=1)
    acled_df['event_date'] = pd.to_datetime(acled_df['event_date'], format='%m/%d/%Y')
    acled_df['year'] = acled_df['event_date'].dt.year
    acled_df['month'] = acled_df['event_date'].dt.month
    acled_df = acled_df[~acled_df['adm2_pcode'].isna()]
    acled_df = acled_df.drop(columns=['event_date'], axis=1)
    acled_df = acled_df.fillna(0)
    acled_df = acled_df.groupby(['year', 'month', 'adm2_pcode'])['fatalities'].agg(['sum','count']).reset_index().rename(columns={'sum': 'fatalities', 'count': 'incidents'})

    # add 1deg neighbouring conflict
    neighbours_1deg_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'neighbouring_adm2_1deg.csv'))
    dates_df = acled_df[['year', 'month']].drop_duplicates()
    neighbours_1deg_df = dates_df.merge(neighbours_1deg_df, how='cross')
    neighbours_1deg_df = neighbours_1deg_df.merge(acled_df, left_on=['year', 'month', 'neighbour'], right_on=['year', 'month', 'adm2_pcode'])
    neighbours_1deg_df = neighbours_1deg_df.drop(['adm2_pcode_y', 'neighbour'], axis=1).rename(columns={'adm2_pcode_x': 'adm2_pcode'})
    neighbours_1deg_df = neighbours_1deg_df.groupby(['year', 'month', 'adm2_pcode']).mean().reset_index().rename(columns={
        'fatalities': 'fatalities_1deg_neighbours',
        'incidents': 'incidents_1deg_neighbours'
        })
    neighbours_1deg_df = neighbours_1deg_df.fillna(0)

    # add 2deg neighbouring conflict
    neighbours_2deg_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'neighbouring_adm2_2deg.csv'))
    dates_df = acled_df[['year', 'month']].drop_duplicates()
    neighbours_2deg_df = dates_df.merge(neighbours_2deg_df, how='cross')
    neighbours_2deg_df = neighbours_2deg_df.merge(acled_df, left_on=['year', 'month', 'neighbour'], right_on=['year', 'month', 'adm2_pcode'])
    neighbours_2deg_df = neighbours_2deg_df.drop(['adm2_pcode_y', 'neighbour'], axis=1).rename(columns={'adm2_pcode_x': 'adm2_pcode'})
    neighbours_2deg_df = neighbours_2deg_df.groupby(['year', 'month', 'adm2_pcode']).mean().reset_index().rename(columns={
        'fatalities': 'fatalities_2deg_neighbours',
        'incidents': 'incidents_2deg_neighbours'
        })
    neighbours_2deg_df = neighbours_2deg_df.fillna(0)
    acled_df = acled_df.merge(neighbours_1deg_df, on=['year', 'month', 'adm2_pcode'], how='outer')
    acled_df = acled_df.merge(neighbours_2deg_df, on=['year', 'month', 'adm2_pcode'], how='outer')

    acled_df = acled_df.rename(columns={'incidents': 'conflict_incidents', 'fatalities': 'conflict_fatalities'})

    master_dataset_df = master_dataset_df.merge(acled_df, on=['year', 'month', 'adm2_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 9: Add IOD IDP
# Datasets used:
# 2014 to 2018: https://dtm.iom.int/sites/g/files/tmzbdl1461/files/datasets/DTM_YEMEN_Round_37_November_2018.xlsx
# 2019: https://dtm.iom.int/sites/g/files/tmzbdl1461/files/datasets/DTM%20Yemen_%20Rapid%20Displacement%20Tracking%20%28RDT%29%20Dataset%20_01%20Jan%20to%2030%20Nov%202019.xlsx
# 2020: https://dtm.iom.int/sites/g/files/tmzbdl1461/files/datasets/Yemen%20-%20Rapid%20Displacement%20Tracking%20Dataset%20%2801%20January%20-%2005%20Dec%202020%29.xlsx
# 2021: https://dtm.iom.int/sites/g/files/tmzbdl1461/files/datasets/20220109_RDT_Weekly%20Update-%2002%20January%202022%20-%2009%20January%202022%20External%20District%20Level.xlsx
# 2022: https://dtm.iom.int/sites/g/files/tmzbdl1461/files/datasets/20230102_RDT%20Dataset-%2025%20December%202022%20-%2031%20December%202022%20External%20District%20Level.xlsx
# 2023: https://dtm.iom.int/sites/g/files/tmzbdl1461/files/datasets/20230704_RDT%20Dataset-%2025%20June%202023%20-%2001%20July%202023%20External%20District%20Level.xlsx
if True:
    print('Adding iom idp')
    iom_folder = os.path.join(DATA_FOLDER, 'iom')

    iom_idp_2014_2018_df = pd.read_excel(os.path.join(iom_folder, 'DTM_YEMEN_Round_37_November_2018.xlsx'), sheet_name='01_1_IDPs Origin', skiprows=1)

    iom_idp_2014_2018_df = iom_idp_2014_2018_df[['Year of Displacement', 'Month of displacement  (2015 - 2016)', 'District PCode', 'District of Orgin PCode', 'Total IDPs HH']]
    iom_idp_2014_2018_df = iom_idp_2014_2018_df.rename(columns={
        'Year of Displacement': 'year',
        'Month of displacement  (2015 - 2016)': 'month',
        'District PCode': 'adm2_pcode',
        'District of Orgin PCode': 'adm2_pcode_from',
        'Total IDPs HH': 'idps'
    })

    iom_idp_2014_2018_df['adm2_pcode'] = 'YE' + iom_idp_2014_2018_df['adm2_pcode'].astype(str)
    iom_idp_2014_2018_df['adm2_pcode_from'] = 'YE' + iom_idp_2014_2018_df['adm2_pcode_from'].astype(str)
    iom_idp_2014_2018_df['adm2_pcode'] = iom_idp_2014_2018_df['adm2_pcode'].replace('YE1604', 'YE1405')
    iom_idp_2014_2018_df['adm2_pcode_from'] = iom_idp_2014_2018_df['adm2_pcode_from'].replace('YE1604', 'YE1405')

    iom_idp_2014_2018_df_to = iom_idp_2014_2018_df[['year', 'month', 'adm2_pcode', 'idps']].rename(columns={'idps': 'displaced_to'}).groupby(['year', 'month', 'adm2_pcode']).sum().reset_index()
    iom_idp_2014_2018_df_from = iom_idp_2014_2018_df[['year', 'month', 'adm2_pcode_from', 'idps']].rename(columns={'idps': 'displaced_from', 'adm2_pcode_from': 'adm2_pcode'}).groupby(['year', 'month', 'adm2_pcode']).sum().reset_index()
    iom_idp_2014_2018_df = iom_idp_2014_2018_df_to.merge(iom_idp_2014_2018_df_from, on=['year', 'month', 'adm2_pcode'], how='outer')

    # 2019-2020
    iom_idp_2019_df = pd.read_excel(os.path.join(iom_folder, 'DTM Yemen_ Rapid Displacement Tracking (RDT) Dataset _01 Jan to 30 Nov 2019.xlsx'), sheet_name='RDT-IDPs', skiprows=1)
    iom_idp_2020_df = pd.read_excel(os.path.join(iom_folder, 'Yemen - Rapid Displacement Tracking Dataset (01 January - 05 Dec 2020).xlsx'), sheet_name='IDPs', skiprows=1)

    iom_idp_2019_df = iom_idp_2019_df[['Displacement Date', 'District', 'District of Origin', 'Total # of HH']]
    iom_idp_2020_df = iom_idp_2020_df[['Displacement Date', 'District', 'District of Origin', 'Total # of HH']]

    iom_idp_2019_2020_df = pd.concat([iom_idp_2019_df, iom_idp_2020_df])

    iom_idp_2019_df = iom_idp_2019_2020_df.rename(columns={
        'Displacement Date': 'date',
        'District': 'adm2_name',
        'District of Origin': 'adm2_name_from',
        'Total # of HH': 'idps'
    })

    iom_idp_2019_df['date'] = pd.to_datetime(iom_idp_2019_df['date'], format='%Y-%m-%d')
    iom_idp_2019_df['year'] = iom_idp_2019_df['date'].dt.year
    iom_idp_2019_df['month'] = iom_idp_2019_df['date'].dt.month
    iom_idp_2019_df = iom_idp_2019_df.drop(columns=['date'])

    iom_idp_pcodes_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'adm2_correction-Grid view.csv'))

    iom_idp_2019_df_to = iom_idp_2019_df.merge(iom_idp_pcodes_df[['adm2_name_wrong', 'adm2_pcode']], left_on=['adm2_name'], right_on=['adm2_name_wrong'], how='left')
    iom_idp_2019_df_to = iom_idp_2019_df_to[['year', 'month', 'adm2_pcode', 'idps']].rename(columns={'idps': 'displaced_to'}).groupby(['year', 'month', 'adm2_pcode']).sum().reset_index()
    iom_idp_2019_df_to['displaced_to'] = iom_idp_2019_df_to['displaced_to'].abs()
    iom_idp_2019_df_from = iom_idp_2019_df.merge(iom_idp_pcodes_df[['adm2_name_wrong', 'adm2_pcode']], left_on=['adm2_name_from'], right_on=['adm2_name_wrong'], how='left')
    iom_idp_2019_df_from = iom_idp_2019_df_from[['year', 'month', 'adm2_pcode', 'idps']].rename(columns={'idps': 'displaced_from'}).groupby(['year', 'month', 'adm2_pcode']).sum().reset_index()
    iom_idp_2019_df_from['displaced_from'] = iom_idp_2019_df_from['displaced_from'].abs()
    iom_idp_2019_df = iom_idp_2019_df_to.merge(iom_idp_2019_df_from, on=['year', 'month', 'adm2_pcode'], how='outer')

    # 2021-2023
    iom_idp_2021_df = pd.read_excel(os.path.join(iom_folder, '20220109_RDT_Weekly Update- 02 January 2022 - 09 January 2022 External District Level.xlsx'), sheet_name='IDPs')
    iom_idp_2021_df = iom_idp_2021_df[['Date of Event', 'Dis_Pcode', 'Dis_Coming_Pcode', 'Total Number Of Households']]
    iom_idp_2021_df = iom_idp_2021_df[iom_idp_2021_df['Date of Event']<='2021-12-31']

    iom_idp_2022_df = pd.read_excel(os.path.join(iom_folder, '20230102_RDT Dataset- 25 December 2022 - 31 December 2022 External District Level.xlsx'), sheet_name='IDPs')
    iom_idp_2022_df = iom_idp_2022_df[['Date of Event', 'Dis_Pcode', 'Dis_Coming_Pcode', 'Total Number Of Households']]

    iom_idp_2023_df = pd.read_excel(os.path.join(iom_folder, '20230704_RDT Dataset- 25 June 2023 - 01 July 2023 External District Level.xlsx'), sheet_name='IDPs')
    iom_idp_2023_df = iom_idp_2023_df[['Date of Event', 'Dis_Pcode', 'Dis_Coming_Pcode', 'Total Number Of Households']]

    iom_idp_2021_2023_df = pd.concat([iom_idp_2021_df, iom_idp_2022_df, iom_idp_2023_df])

    iom_idp_2021_2023_df = iom_idp_2021_2023_df.rename(columns={
        'Date of Event': 'date',
        'Dis_Pcode': 'adm2_pcode',
        'Dis_Coming_Pcode': 'adm2_pcode_from',
        'Total Number Of Households': 'idps'
    })

    iom_idp_2021_2023_df['date'] = pd.to_datetime(iom_idp_2021_2023_df['date'], format='%Y-%m-%d')
    iom_idp_2021_2023_df['year'] = iom_idp_2021_2023_df['date'].dt.year
    iom_idp_2021_2023_df['month'] = iom_idp_2021_2023_df['date'].dt.month
    iom_idp_2021_2023_df = iom_idp_2021_2023_df.drop(columns=['date'])

    iom_idp_2021_2023_df_to = iom_idp_2021_2023_df[['year', 'month', 'adm2_pcode', 'idps']].groupby(['year', 'month', 'adm2_pcode']).sum().reset_index().rename(columns={'idps': 'displaced_to'})
    iom_idp_2021_2023_from = iom_idp_2021_2023_df[['year', 'month', 'adm2_pcode_from', 'idps']].groupby(['year', 'month', 'adm2_pcode_from']).sum().reset_index().rename(columns={'idps': 'displaced_from', 'adm2_pcode_from': 'adm2_pcode'})
    iom_idp_2021_2023_df = iom_idp_2021_2023_from.merge(iom_idp_2021_2023_df_to, on=['year', 'month', 'adm2_pcode'], how='outer')

    # add all idp iom datafames
    iom_idp_df = pd.concat([iom_idp_2014_2018_df, iom_idp_2019_df, iom_idp_2021_2023_df])
    iom_idp_df = iom_idp_df[~iom_idp_df['adm2_pcode'].isin(['YE1319', 'YE1324'])]
    iom_idp_df = iom_idp_df.sort_values(by=['year', 'month', 'adm2_pcode'], ascending=True)
    master_dataset_df = master_dataset_df.merge(iom_idp_df, on=['year', 'month', 'adm2_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 10: Add World Bank market prices
# https://microdata.worldbank.org
# Currently, the world bank split this dataset into 3 separate datasets, one for food prices, one for fuel prices and one for exchange rate.
# When we modeled, we received a combined version, which is the csv file we load here.
# the three seperate datsets can be found at:
# food prices: https://microdata.worldbank.org/index.php/api/tables/data/fcv/wld_2021_rtfp_v02_m
# fuel prices: https://microdata.worldbank.org/index.php/api/tables/data/FCV/WLD_2023_RTEP_v01_M
# exchange rate: https://microdata.worldbank.org/index.php/api/tables/data/FCV/WLD_2023_RTFX_v01_M
if True:
    print('Adding world bank food prices')
    world_bank_market_prices_df = pd.read_csv(os.path.join(DATA_FOLDER, 'world_bank_market_prices.csv'))
    world_bank_market_prices_df = world_bank_market_prices_df[[
        'year',
        'month',
        'adm1_name',
        'c_beans',
        'c_eggs',
        'c_exchange_rate_unofficial',
        'c_fuel_diesel',
        'c_fuel_gas',
        'c_fuel_petrol_gasoline',
        'c_lentils',
        'c_livestock_sheep_two_year_old_male',
        'c_millet',
        'c_oil',
        'c_onions',
        'c_peas',
        'c_potatoes',
        'c_rice',
        'c_salt',
        'c_sorghum',
        'c_sugar',
        'c_tomatoes',
        'c_wheat',
        'c_wheat_flour'
    ]]
    world_bank_market_prices_df = world_bank_market_prices_df[world_bank_market_prices_df['adm1_name']!='Market Average']
    world_bank_market_prices_pcodes_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'world_bank_adm1_pcodes.csv'))
    world_bank_market_prices_df = world_bank_market_prices_df.merge(world_bank_market_prices_pcodes_df, on='adm1_name', how='left')
    world_bank_market_prices_df = world_bank_market_prices_df.drop(columns=['adm1_name'], axis=1)
    world_bank_market_prices_df = world_bank_market_prices_df.groupby(['year', 'month', 'adm1_pcode']).mean().reset_index()
    master_dataset_df = master_dataset_df.merge(world_bank_market_prices_df, on=['year', 'month', 'adm1_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]


def zscore(df, column, new_column_name, method='mean'):
    if method == 'mean':
        df[new_column_name] = (df[column] - df[column].mean())/df[column].std()
    elif method == 'median':
        df[new_column_name] = (df[column] - df[column].median())/df[column].std()
    else:
        print(f'Unknown method: {method}')
    return df


# Step 11: Add WFP NDVI
# https://data.humdata.org/dataset/7907439a-0138-4b8e-a4cf-4f0233e0d694/resource/5a30f36a-aa4c-4261-a24b-22bda7639e4c/download/yem-ndvi-adm2-full.csv
if True:
    print('Adding ndvi')
    ndvi_df = pd.read_csv(os.path.join(DATA_FOLDER, 'yem-ndvi-adm2-full.csv')).iloc[1:, :].rename(columns={'ADM2_PCODE': 'adm2_pcode'})
    ndvi_df = ndvi_df[~ndvi_df['date'].isna()]
    ndvi_df['date'] = pd.to_datetime(ndvi_df['date'], format='%m/%d/%Y')
    
    ndvi_df['year'] = ndvi_df['date'].dt.year
    ndvi_df['month'] = ndvi_df['date'].dt.month
    ndvi_df = ndvi_df.drop(columns=['date', 'adm2_id', 'n_pixels', 'viq'], axis=1)
    ndvi_df = ndvi_df.rename(columns={'vim': 'ndvi_average', 'vim_avg': 'ndvi_lt_avg'})
    ndvi_df['ndvi_average'] = ndvi_df['ndvi_average'].astype(float)
    ndvi_df['ndvi_lt_avg'] = ndvi_df['ndvi_lt_avg'].astype(float)
    ndvi_df = ndvi_df.groupby(['year', 'month', 'adm2_pcode']).mean().reset_index()
    ndvi_df['ndvi_anomaly'] = ndvi_df['ndvi_average'].divide(ndvi_df['ndvi_lt_avg']).fillna(0)

    # add z-scores
    ndvi_new_df = pd.DataFrame()
    grouper = ndvi_df.groupby(['month', 'adm2_pcode'])
    for name, ndvi_month_and_adm2_df in grouper:
        ndvi_month_and_adm2_df = zscore(ndvi_month_and_adm2_df, 'ndvi_average', 'ndvi_zscore_with_mean', method='mean')
        ndvi_month_and_adm2_df = zscore(ndvi_month_and_adm2_df, 'ndvi_average', 'ndvi_zscore_with_median', method='median')
        ndvi_new_df = pd.concat([ndvi_new_df, ndvi_month_and_adm2_df])
    master_dataset_df = master_dataset_df.merge(ndvi_new_df, on=['year', 'month', 'adm2_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 12: Add WFP Rainfall
# https://data.humdata.org/dataset/902bd76c-3f3c-42ad-8547-f9c31deb7937/resource/8ebd5130-715e-4bc9-920c-773d461b0754/download/yem-rainfall-adm2-full.csv
if True:
    print('Adding rainfall')
    rainfall_df = pd.read_csv(os.path.join(DATA_FOLDER, 'yem-rainfall-adm2-full.csv')).iloc[1:, :].rename(columns={'ADM2_PCODE': 'adm2_pcode'})
    rainfall_df['date'] = pd.to_datetime(rainfall_df['date'], format='%m/%d/%Y')
    rainfall_df['year'] = rainfall_df['date'].dt.year
    rainfall_df['month'] = rainfall_df['date'].dt.month
    rainfall_df = rainfall_df.drop(columns=['date', 'adm2_id', 'n_pixels', 'r1h', 'r1h_avg', 'r3h', 'r3h_avg', 'rfq', 'r1q', 'r3q', 'version'], axis=1)

    rainfall_df = rainfall_df.rename(columns={'rfh': 'rainfall_monthly_total', 'rfh_avg': 'rainfall_monthly_lt_average'})
    rainfall_df['rainfall_monthly_total'] = rainfall_df['rainfall_monthly_total'].astype(float)
    rainfall_df['rainfall_monthly_lt_average'] = rainfall_df['rainfall_monthly_lt_average'].astype(float)
    rainfall_df = rainfall_df.groupby(['year', 'month', 'adm2_pcode']).mean().reset_index()
    rainfall_df['rainfall_anomaly'] = rainfall_df['rainfall_monthly_total'].divide(rainfall_df['rainfall_monthly_lt_average']).fillna(0)

    # add z-scores
    rainfall_new_df = pd.DataFrame()
    grouper = rainfall_df.groupby(['month', 'adm2_pcode'])
    for name, rainfall_month_and_adm2_df in grouper:
        rainfall_month_and_adm2_df = zscore(rainfall_month_and_adm2_df, 'rainfall_monthly_total', 'rainfall_zscore_with_mean', method='mean')
        rainfall_month_and_adm2_df = zscore(rainfall_month_and_adm2_df, 'rainfall_monthly_total', 'rainfall_zscore_with_median', method='median')
        rainfall_new_df = pd.concat([rainfall_new_df, rainfall_month_and_adm2_df])
    master_dataset_df = master_dataset_df.merge(rainfall_new_df, on=['year', 'month', 'adm2_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 13: Add FAO SPI
# mailed by FAO
if True:
    print('Adding spi')
    spi_df = pd.read_csv(os.path.join(DATA_FOLDER, 'dump_ye_rainfall_and_spi_data.csv'))
    spi_df = spi_df.drop(columns=['district', 'mean_prec', 'sum_prec', 'max_prec', 'min_prec', 'std_prec', 'cnt_prec'], axis=1)
    for c in ['spi1', 'spi3', 'spi6', 'spi12', 'spi24']:
        spi_df[c] = spi_df[c].round(9)
    master_dataset_df = master_dataset_df.merge(spi_df, left_on=['year', 'month', 'adm2_pcode'], right_on=['year', 'month', 'pcode_id'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 14: Add Crop calendar
# mailed by FAO
if True:
    print('Adding crop calendar')
    agri_zone_type_df = pd.read_csv(os.path.join(DATA_FOLDER, 'lookups', 'agri_zone_type.csv'))


    growing_months_df = pd.read_excel(os.path.join(DATA_FOLDER, 'AgricultualCalendar_Yemen_V2.xlsx'), sheet_name='growing months')
    growing_months_df = pd.melt(growing_months_df, id_vars='agri_zone_type', value_vars=[x for x in range(1, 13)], var_name='month', value_name='is_crop_growing')
    agri_zone_type_df = agri_zone_type_df.merge(growing_months_df, on='agri_zone_type', how='left')
    master_dataset_df = master_dataset_df.merge(agri_zone_type_df, on=['month', 'adm2_pcode'], how='left')
    assert number_of_rows == master_dataset_df.shape[0]

# Step 16: Write to csv
master_dataset_df = master_dataset_df.sort_values(by=['year', 'month', 'adm2_pcode'], ascending=True)
master_dataset_df.to_csv(os.path.join(DATA_FOLDER, 'yemen_jmr_master_dataset_model_repro.csv'), index=False)
