from init_model import get_escalations
import math
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from stargazer.stargazer import Stargazer
import statsmodels.api as sm

from indicator_methods import calculate_rolling_average, filter_solutions, get_method, sort_method_results
from init_model import get_escalations
from scoring_metrics import get_brier_and_r_squared, get_metrics_results
from indicator_threshold_model import binarize_indicator, finding_optimal_thresholds


def trim_solution_space(df, candidates_df, valid_indicators, target_indicator, top_N=5, escalations_only=True, date_column='date', location_column='adm2_pcode'):
    valid_df = filter_solutions(candidates_df, valid_indicators, top_N=top_N)
    df = df.copy()
    all_df = df[[date_column, location_column, target_indicator, f'{target_indicator}_escalations']].copy()
    all_df = sort_method_results(all_df, date_column=date_column, location_column=location_column)
    if escalations_only:
        all_df = get_escalations(all_df, target_indicator=target_indicator)
    all_df = all_df[all_df[target_indicator]>=0]
    for _row_idx, row in valid_df.iterrows():
        m = get_method(row['Method'])
        f = row['Risk Indicator']
        if row['Method'] == 'rsi_of_ema':
            time_window = row['Window Length']
            ema_time_window = int(time_window.split('[')[1].split(',')[0])
            rsi_time_window = int(time_window.split(']')[0].split(', ')[1])
            df = m(df, f, window_length=ema_time_window, window_length2=rsi_time_window, date_column=date_column, location_column=location_column)
        else:
            df = m(df, f, window_length=row['Window Length'], date_column=date_column, location_column=location_column)
        if escalations_only:
            df = get_escalations(df, target_indicator=target_indicator)
        df = df[df[target_indicator]>=0]

        x = binarize_indicator(df[f'_indicator_{f}'], boundary=row['Boundary'], rules_reversed=bool(row['Rules reversed']))
        new_column_name = f'{row["Risk Indicator"]}_{row["Method"]}_{row["Window Length"]}'
        # new_column_name = f'{row["Risk Indicator"]}__{row["Method"]}__{"".join(["zero", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten"][int(i)] for i in str(row["Window Length"]))}'
        all_df[new_column_name] = x
    return valid_df, all_df


def get_correlation_matrix(df, features, target_indicator, escalations_only=True, mode='alert', date_column='date', location_column='adm2_pcode', bin=True):
    '''
        Return the correlation matrix for the given list of features
    '''
    df = combined_model_init_all(df, features, target_indicator=target_indicator, escalations_only=escalations_only, mode=mode, date_column=date_column, location_column=location_column)
    if bin:
        feature_names = [f'_indicator_{f["indicator"]}_bin' for f in features]
    else:
        feature_names = [f'_indicator_{f["indicator"]}' for f in features]
    df_corr = df[feature_names] 
    df_corr.columns = [f['name'] for f in features]
    corr = df_corr.corr()
    return corr


def plot_kernel_density(df, features, target_indicator, escalations_only=True, mode='alert', date_column='date', location_column='adm2_pcode', nr_rows=3, nr_cols=2):
    df = combined_model_init_all(df, features, target_indicator=target_indicator, escalations_only=escalations_only, mode=mode, date_column=date_column, location_column=location_column)    
    _fig, axes = plt.subplots(nrows=nr_rows, ncols=nr_cols, figsize=(10, 12), gridspec_kw={'hspace':0.5})
    for idx, f in enumerate(features):
        graph = sns.kdeplot(data=df, x=f'_indicator_{f["indicator"]}', ax=axes[idx%nr_rows, idx%nr_cols], bw_adjust=0.5)
        axes[idx%nr_rows, idx%nr_cols].title.set_text(f['name'])
        axes[idx%nr_rows, idx%nr_cols].set_title(f['name'], fontdict={'fontsize': 10})
        axes[idx%nr_rows, idx%nr_cols].set_xlabel(f'{f["method_name"]}')
        graph.axvline(f['alert_threshold'], color='orange', label='Heightened threshold')
        graph.axvline(f['alarm_threshold'], color='red', label='Critical threshold')
        if f['rules_reversed']:
            axes[idx%nr_rows, idx%nr_cols].invert_xaxis()
    plt.legend(loc='best')
    plt.show()


def combined_model_init_all(df, features, target_indicator, escalations_only=True, mode='alert', date_column='date', location_column='adm2_pcode', keep_only_target_indicators=True):
    df = df.copy()
    if mode == 'alert':
        threshold_column = 'alert_threshold'
    else:
        threshold_column = 'alarm_threshold'
    for c in features:
        f = c['indicator']
        method = c['method']
        window_length = c['window_length']
        rules_reversed = c['rules_reversed']
        boundary = c[threshold_column]
        if isinstance(window_length, list):
            df = method(df, f, window_length=window_length[0], window_length2=window_length[1], date_column=date_column, location_column=location_column)
        else:
            df = method(df, f, window_length=window_length, date_column=date_column, location_column=location_column)
        df[f'_indicator_{f}_bin'] = binarize_indicator(df[f'_indicator_{f}'], boundary=boundary, rules_reversed=rules_reversed)
    if escalations_only:
        df = get_escalations(df, target_indicator=target_indicator)
    if keep_only_target_indicators:
        df = df[df[target_indicator]>=0]
    return df


def majority_vote(df, features, target_indicator, escalations_only=True, mode='alert', weight=0.5, date_column='date', location_column='adm2_pcode'):
    '''
        Runs a simple majority vote, keep adding one feature at a time and determine the score.
        Visualizes the result
    '''
    df = combined_model_init_all(df, features, target_indicator=target_indicator, escalations_only=escalations_only, mode=mode, date_column=date_column, location_column=location_column)
    method = 'Balanced Accuracy'
    # determine the best order
    scores = []
    for f in features:
        f = f'_indicator_{f["indicator"]}_bin'
        s = get_metrics_results(df[target_indicator], df[f], weight=weight)
        scores.append([f, s[method]])
    best_order = [x for x in sorted(scores, key=lambda x:x[1], reverse=True)]

    # plot the majority vote
    scores = []
    for idx in range(1, len(best_order)+1):
        cutoff = idx/2
        features_to_use = best_order[:idx]
        df['majority_vote'] = 0
        for f in features_to_use:
            df['majority_vote'] += df[f[0]]
        y_score = np.where(df['majority_vote'] >= cutoff, 1, 0)
        result = get_metrics_results(df[target_indicator], y_score, weight=weight)
        scores.append(result[method])
    plt.plot([x for x in range(1, len(best_order)+1)], scores)
    plt.xlabel('Number of indicators')
    plt.ylabel('Balanced Accuracy')
    plt.show()


def check_mode(df, features, mode, target_indicator, escalations_only=True, date_column='date', location_column='adm2_pcode', bin=True, keep_only_target_indicators=True):
    if mode in ['alert', 'alarm']:
        df, features = return_df_and_features(df, features, target_indicator=target_indicator, escalations_only=escalations_only, mode=mode, bin=bin, date_column=date_column, location_column=location_column, keep_only_target_indicators=keep_only_target_indicators)
    else:
        df_alert, features_alert = return_df_and_features(df, features, target_indicator=target_indicator, escalations_only=escalations_only, mode='alert', bin=bin, date_column=date_column, location_column=location_column, keep_only_target_indicators=keep_only_target_indicators)
        df_alarm, features_alarm = return_df_and_features(df, features, target_indicator=target_indicator, escalations_only=escalations_only, mode='alarm', bin=bin, date_column=date_column, location_column=location_column, keep_only_target_indicators=keep_only_target_indicators)
        df = df_alert.merge(df_alarm, on=[date_column, location_column], how='inner')
        for c in df.columns:
            if c.endswith('_x'):
                df = df.rename(columns={c: f'{c.split("_x")[0]}'})
            elif c.endswith('_y'):
                df = df.drop([c], axis=1)

        features = features_alert + features_alarm
    return df, features


def return_df_and_features(df, features, target_indicator, escalations_only=True, mode='alert', bin=True, date_column='date', location_column='adm2_pcode', keep_only_target_indicators=True):
    df = combined_model_init_all(df, features, target_indicator=target_indicator, escalations_only=escalations_only, mode=mode, date_column=date_column, location_column=location_column, keep_only_target_indicators=keep_only_target_indicators)
    if bin:
        to_change = {f'_indicator_{f["indicator"]}_bin': f'{mode}_{f["indicator"]}' for f in features}
    else:
        to_change = {f'_indicator_{f["indicator"]}': f'{mode}_{f["indicator"]}' for f in features}
    df = df.rename(columns=to_change)
    features = list(to_change.values())        
    return df, features


# GLM
def run_glm(df, features, target_indicator, mode='alert', escalations_only=True, date_column='date', location_column='adm2_pcode', bin=True, weight=0.5, interaction_term=False):
    df, features = check_mode(df, features, mode=mode, target_indicator=target_indicator, escalations_only=escalations_only, date_column=date_column, location_column=location_column, bin=bin)
    results, best_order, glm_result = glm(df, features, target_indicator=target_indicator, bin=bin, weight=weight, interaction_term=interaction_term)
    return results, best_order, glm_result, df


def glm(df, features, target_indicator, weight=0.5, bin=True):
    if not bin:
        for f in features:
            try:
                df[f] = (df[f] - df[f].min()) / (df[f].max() - df[f].min())
            except Exception:
                pass

    formula = f'{target_indicator} ~ '
    formula += ''.join([f'{f} + ' for f in features])
    formula = formula[0:-3]
    model = sm.formula.glm(formula, data=df, family=sm.families.Binomial())
    glm_result = model.fit()

    features2 = glm_result.params.index.tolist()
    coefficients = glm_result.params
    best_order = []
    for f, c in zip(features2, coefficients):
        if f != 'Intercept':
            best_order.append([f, round(c, 2)])
            # print(f'    {f}, {c}')
            # try:
            #     best_order.append([f, round(math.exp(c), 2)])
            # except OverflowError:
            #     best_order.append([f, 1000])
    best_order = [[x[0], x[1]] for x in sorted(best_order, key=lambda x:x[1], reverse=True)]
    
    df['predict'] = glm_result.predict()
    all_scores = finding_optimal_thresholds(df[target_indicator], df['predict'], weight=weight, lower_boundary=0, upper_boundary=1)
    boundary = all_scores[weight][0][0]
    y_score = binarize_indicator(df['predict'], boundary)
    results = get_metrics_results(df[target_indicator], y_score)

    return results, best_order, glm_result, df


def run_rfe_on_glm(df, features, target_indicator, mode='alert', escalations_only=True, date_column='date', location_column='adm2_pcode', bin=True, weight=0.5, plot=True, interaction_term=False, interactions=[], set_number_of_indicators=False, meta_indicator=False, indicators_to_keep=[]):
    df, features = check_mode(df, features, target_indicator=target_indicator, mode=mode, escalations_only=escalations_only, date_column=date_column, location_column=location_column, bin=bin)
    
    # add meta indicator if needed
    if meta_indicator:
        alert_features = [f for f in features if f.startswith('alert')]
        alarm_features = [f for f in features if f.startswith('alarm')]
        if len(alert_features) > 0:
            df['_alert_meta_indicator'] = df[alert_features].sum(axis=1)
            # rolling average of meta indicator
            df = calculate_rolling_average(df, '_alert_meta_indicator', window_length=4, date_column=date_column, location_column=location_column)
            features.append('_indicator__alert_meta_indicator')
        if len(alarm_features) > 0:
            df['_alarm_meta_indicator'] = df[alarm_features].sum(axis=1)
            # rolling average of meta indicator
            df = calculate_rolling_average(df, '_alarm_meta_indicator', window_length=4, date_column=date_column, location_column=location_column)
            features.append('_indicator__alarm_meta_indicator')
        


    # add interaction term if needed
    if not isinstance(interaction_term, bool):
        for interaction_on in interactions:
            features += [f'{f}:{interaction_term}' for f in features if interaction_on in f]
    
        if interaction_term not in df.columns.tolist():
            df[interaction_term] = df[f'{interaction_term}_x'].copy()

    best_result, best_order_all_positives, best_glm_result, df_new = rfe_on_glm(df, features, target_indicator=target_indicator, bin=bin, weight=weight, plot=plot, set_number_of_indicators=set_number_of_indicators, indicators_to_keep=indicators_to_keep)
    scores = return_scores_glm(df_new, best_glm_result, target_indicator=target_indicator, weight=weight)
    # stargazer = Stargazer([best_glm_result])
    return df_new, best_glm_result, scores


def rfe_on_glm(df, features, target_indicator, bin=True, weight=0.5, plot=True, set_number_of_indicators=False, indicators_to_keep=[]):
    method = 'Balanced Accuracy'
    scores = []
    best_found = False
    best_order_all_positives = []
    best_glm_result = []
    best_result = []
    best_df = []
    best_score = 0
    while len(features) > 0:
        # run the GLM
        # print(f'Number of features = {len(features)}')
        result, best_order, glm_result, df_new = glm(df.copy(), features, target_indicator=target_indicator, weight=weight, bin=bin)
        scores.append([len(best_order), result[method]])

        if best_score < result[method]:
            if isinstance(set_number_of_indicators, bool):
                best_score = result[method]
                best_order_all_positives = best_order
                best_glm_result = glm_result
                best_result = result
                best_df = df_new
        if not isinstance(set_number_of_indicators, bool):
            if len(features) == set_number_of_indicators:
                best_order_all_positives = best_order
                best_glm_result = glm_result
                best_result = result
                best_df = df_new

        if len(indicators_to_keep) == 0:
            pvalues = glm_result.pvalues.iloc[1:]
            worst_performing_feature = pvalues.idxmax()
        else:
            # pick the least performing one outside the indicators_to_keep. When there are only indicators left that are in this list, it is ko to remove them
            # print(len(features), len(indicators_to_keep))
            if len(indicators_to_keep) >= len(features):
                pvalues = glm_result.pvalues.iloc[1:]
                worst_performing_feature = pvalues.idxmax()
            else:
                pvalues = glm_result.pvalues.iloc[1:].sort_values(ascending=False)
                success = False
                idx = 0
                while not success:
                    worst_performing_feature = pvalues.index[idx]
                    # print(worst_performing_feature)
                    if worst_performing_feature not in indicators_to_keep:
                        success = True
                    idx += 1

        # drop least performing feature
        # worst_performing_feature = best_order[-1][0]
        # print(f'Removing feature {worst_performing_feature}')
        features = [f for f in features if f != worst_performing_feature]
    
    if plot:
        plt.plot([s[0] for s in scores], [s[1] for s in scores])
        plt.xlabel('Number of indicators')
        plt.ylabel('Balanced Accuracy')
        plt.show()
    return best_result, best_order_all_positives, best_glm_result, best_df


def return_scores_glm(df, glm_result, target_indicator, weight=0.5):
    y_real = df[target_indicator]
    y_score = glm_result.predict()
    BS, r2 = get_brier_and_r_squared(y_real, y_score)
    # print(f'Brier: {BS}, RS: {r2}')
    all_scores = finding_optimal_thresholds(df[target_indicator], df['predict'], weight=weight, lower_boundary=0, upper_boundary=1)
    boundary = all_scores[weight][0][0]
    y_score = binarize_indicator(df['predict'], boundary)
    results = get_metrics_results(df[target_indicator], y_score)
    # print(f'Loss w=1/2: {results["Loss"]}')
    # print(f'Loss w=1/3: {results["Loss (w=1/3)"]}')
    # print(f'Loss w=2/3: {results["Loss (w=2/3)"]}')
    return {'Brier': BS, 'RS': r2, 'Loss w=1/2': results['Loss'], 'Loss w=1/3': results['Loss (w=1/3)'], 'Loss w=2/3': results['Loss (w=2/3)']}