import numpy as np
import pandas as pd


def sort_method_results(df, date_column, location_column):
    df = df.sort_values(by=[date_column, location_column], ascending=True)
    return df


def calculate_rolling_average(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    """
        Calculates the Rolling Average of the column 'f' over the last 'window_lenght' number of months.
        Input:
            df: Dataframe
            f:  column name over which you want to calculate the EMA
            window_lenght: number of rows (months) you would like to use when calculating the EMA
        Returns df with an extra column called '_indicator_name_of_indicator'
    """
    df = df.sort_values(by=[date_column], ascending=True)
    df[f'_indicator_{f}'] = df.groupby([location_column])[f].transform(lambda x: x.rolling(window_length, 1).mean().shift().bfill()).tolist()
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_perc_of_rolling_average(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    """
        Calculates the percentage of the current months value for 'f' with respect to the Rolling Average of the column 'f' over the last 'window_lenght' number of months.
        Input:
            df: Dataframe
            f:  column name over which you want to calculate the EMA
            window_lenght: number of rows (months) you would like to use when calculating the EMA
        Returns df with an extra column called '_indicator_name_of_indicator'
    """
    df = df.sort_values(by=[date_column], ascending=True)
    df['rolling_average'] = df.groupby([location_column])[f].transform(lambda x: x.rolling(window_length, 1).mean().shift().bfill()).tolist()
    df[f'_indicator_{f}'] = df[f].divide(df['rolling_average'])
    df[f'_indicator_{f}'] = np.select([df['rolling_average'] == 0, df[f] == 0, True], [1, 0, df[f'_indicator_{f}']])
    df = df.drop(['rolling_average'], axis=1)
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_ema(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    """
        Calculates the Exponential Moving Average of the column 'f' over the last 'window_lenght' number of months.
        Input:
            df: Dataframe
            f:  column name over which you want to calculate the EMA
            window_lenght: number of rows (months) you would like to use when calculating the EMA
        Returns df with an extra column called '_indicator_name_of_indicator'
    """
    df = df.sort_values(by=[date_column], ascending=True)
    df[f'_indicator_{f}'] = df.groupby([location_column])[f].transform(lambda x: x.ewm(alpha=2/(window_length+1)).mean()).tolist()
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_perc_of_ema(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    """
        Calculates the percentage of the current months value for 'f' with respect to the Exponential Moving Average of the column 'f' in the last 'window_length' number of months.
        Input:
            df: Dataframe
            f:  column name over which you want to calculate the EMA
            window_lenght: number of rows (months) you would like to use when calculating the EMA
        Returns df with an extra column called '_indicator_name_of_indicator'
    """
    df = df.sort_values(by=[date_column], ascending=True)
    df['ema'] = df.groupby([location_column])[f].transform(lambda x: x.ewm(alpha=2/(window_length+1)).mean().shift().bfill()).tolist()
    df[f'_indicator_{f}'] = df[f].divide(df['ema'])
    df[f'_indicator_{f}'] = np.select([df['ema'] == 0, df[f] == 0, True], [1, 0, df[f'_indicator_{f}']])
    df = df.drop(['ema'], axis=1)
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_rsi(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    """
        Calculates the Relative Strenght Index over column 'f'
        Input:  
            df: Dataframe
            f: column name over which you want to calculate the rsi
            window_lenght: number of rows (months) you would like to use when calculating the RSI
        Returns df with an extra column called '_indicator_name_of_indicator'
    """
    df = df.sort_values(by=[location_column, date_column], ascending=True)
    df['diff'] = df.groupby([location_column])[f].diff()
    # Make two series: one for lower closes and one for higher closes
    df['up'] = df['diff'].clip(lower=0)
    df['down'] = df['diff'].clip(upper=0).abs()

    df['ma_up'] = df.groupby([location_column])['up'].transform(lambda x: x.ewm(alpha=1/window_length).mean()).tolist()
    df['ma_down'] = df.groupby([location_column])['down'].transform(lambda x: x.ewm(alpha=1/window_length).mean()).tolist()

    df['rs'] = df['ma_up'] / df['ma_down']
    df['rs'] = np.select([(df['ma_down'] == 0) & (df['ma_up']== 0), df['ma_down'] == 0, df['ma_up'] == 0, True], [0, 100, 0, df['rs']])
    df[f'_indicator_{f}'] = 100 - (100/(1 + df['rs']))
    df[f'_indicator_{f}'] = df[f'_indicator_{f}'].fillna(0)
    df = df.drop(['diff', 'up', 'down', 'ma_up', 'ma_down', 'rs'], axis=1)
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_rsi_of_ema(df, f, window_length, window_length2=14, date_column='date', location_column='adm2_pcode', use_crop_calendar=False):
    '''
        Returns the Relative Strength Index of the Exponential Moving Average of the column 'f', where the RSI is determined over the last 'window-length' number of months
        The number of months used in the calculation of the EMA must be changed manually.
        Input:  
            df: Dataframe
            f: column name over which you want to calculate the rsi of ema
            window_lenght: number of rows (months) you would like to use when calculating the EMA
            window_length2: number of rows (months) you would like to use when calculating the RSI
        Returns df with an extra column called '_indicator_name_of_indicator'
    '''
    df = calculate_ema(df, f, window_length, date_column, location_column)
    df = df.rename(columns={f'_indicator_{f}': f'{f}_ema'})
    df = calculate_rsi(df, f'{f}_ema', window_length2, date_column, location_column)
    df = df.drop([f'{f}_ema'], axis=1)
    df = df.rename(columns={f'_indicator_{f}_ema': f'_indicator_{f}'})
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_spi(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    df[f'_indicator_{f}'] = df[f]
    if use_crop_calendar:
        df[f'_indicator_{f}'] = np.select([df['is_crop_growing'] == 0, df['is_crop_growing'] == 1], [1, df[f'_indicator_{f}']])
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_nothing(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    df[f'_indicator_{f}'] = df[f]
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_rolling_average_of_zscore(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    df = calculate_rolling_average(df, f, window_length=window_length, date_column=date_column, location_column=location_column)
    if use_crop_calendar:
        df[f'_indicator_{f}'] = np.select([df['is_crop_growing'] == 0, df['is_crop_growing'] == 1], [1, df[f'_indicator_{f}']])
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_ema_of_zscore(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    df = calculate_ema(df, f, window_length=window_length, date_column=date_column, location_column=location_column)
    if use_crop_calendar:
        df[f'_indicator_{f}'] = np.select([df['is_crop_growing'] == 0, df['is_crop_growing'] == 1], [1, df[f'_indicator_{f}']])
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def calculate_average_of_anomaly_of_long_term_average(df, f, window_length, date_column, location_column, use_crop_calendar=False):
    """
        Returns the average anomaly % for the last window_length numnber of months.
    """
    df = df.sort_values(by=[location_column, date_column], ascending=True)
    df[f'_indicator_{f}'] = df[f].copy()
    for n in range(1, window_length):
        df[f'_indicator_{f}'] += df.groupby(location_column)[f].shift(n)
    df[f'_indicator_{f}'] = df[f'_indicator_{f}'].divide(window_length)
    if use_crop_calendar:
        df[f'_indicator_{f}'] = np.select([df['is_crop_growing'] == 0, df['is_crop_growing'] == 1], [1, df[f'_indicator_{f}']])
    df = sort_method_results(df, date_column=date_column, location_column=location_column)
    return df


def determine_lower_and_upper_boundaries(method, rules_reversed=False, set_boundaries=False, lower_boundary=0, upper_boundary=1):
    if set_boundaries:
        return lower_boundary, upper_boundary
    if method in [calculate_rsi, calculate_rsi_of_ema]:
        return 0, 100
    elif method in [calculate_rolling_average, calculate_ema]:
        return 0, 100
    elif method in [calculate_perc_of_rolling_average, calculate_perc_of_ema]:
        if not rules_reversed:
            return 1, 2
        else:
            return 0, 1
    elif method in [calculate_spi]:
        return -4, 0
    elif method in [calculate_rolling_average_of_zscore, calculate_ema_of_zscore]:
        return -3, 3
    elif method in [calculate_average_of_anomaly_of_long_term_average]:
        return 0, 1
    elif method in [calculate_nothing]:
        return 0, 5
    else:
        print(f'ERROR!!! Unknown method {method}')


def get_method(method_name):
    method_match = {
        'rsi': calculate_rsi,
        'rsi_of_ema': calculate_rsi_of_ema,
        'rolling_average': calculate_rolling_average,
        'ema': calculate_ema,
        'perc_of_rolling_average': calculate_perc_of_rolling_average,
        'perc_of_ema': calculate_perc_of_ema,
        'spi': calculate_spi,
        'rolling_average_of_zscore': calculate_rolling_average_of_zscore,
        'ema_of_zscore': calculate_ema_of_zscore,
        'average_of_anomaly_of_long_term_average': calculate_average_of_anomaly_of_long_term_average,
        'nothing': calculate_nothing
    }
    return method_match[method_name]


def filter_solutions(df, valid_indicators, top_N=5):
    if len(valid_indicators) > 0:
        df = df[df['Risk Indicator'].isin(valid_indicators)]
    valid_df = pd.DataFrame(columns=df.columns)

    for row_idx, row in df.iterrows():
        add = True

        if row['Loss'] >= 0.5:
            add = False

        if row['FPR'] >= 0.5:
            add = False

        if row['FNR'] > 3 * row['FPR']:
            add = False
        if row['FPR'] > 2 * row['FNR']:
            add = False

        if row['Method'] == 'rsi_of_ema':
            time_window = row['Window Length']
            ema_time_window = int(time_window.split('[')[1].split(',')[0])
            rsi_time_window = int(time_window.split(']')[0].split(', ')[1])
            if ema_time_window + rsi_time_window <= 5:
                add = False

        if row['Method'] in ['rsi', 'rsi_of_ema']:
            if row['Boundary'] <= 50:
                add = False

        if row['Method'] in ['rolling_average_of_zscore', 'ema_of_zscore']:
            if row['Boundary'] >= 0.5:
                add = False

        if add:
            valid_df.loc[len(valid_df.index)] = row.tolist()

    valid_df = valid_df.sort_values(by=['Loss'], ascending=True)
    valid_df = valid_df.groupby(['Risk Indicator', 'Method']).head(top_N).reset_index(drop=True)
    return valid_df