from indicator_methods import determine_lower_and_upper_boundaries
from init_model import get_escalations
import numpy as np
import pandas as pd
import plotly as plotly
import plotly.graph_objs as go
from scoring_metrics import get_fpr_and_fnr, get_loss, get_metrics_results


def run_all(df, columns, methods, time_windows, target_indicator, escalations_only, rules_reversed=False, weight=0.5, plot=True, keep_all=True, use_crop_calendar=False):
    if not isinstance(columns, list):
        columns = [columns]
    if not isinstance(methods, list):
        methods = [methods]
    if not isinstance(time_windows, list):
        time_windows = [time_windows]

    df_metric_results = pd.DataFrame()

    for method in methods:
        for column in columns:
            for time_window in time_windows:
                print(method.__name__.split(f'calculate_')[1], column, time_window)
                metric_results = run_specific(df, column, method, window_length=time_window, target_indicator=target_indicator, escalations_only=escalations_only, rules_reversed=rules_reversed, weight=weight, plot=plot, keep_all=keep_all, use_crop_calendar=use_crop_calendar)
                df_metric_results = pd.concat([df_metric_results, metric_results])
    return df_metric_results


def run_specific(df, f, method, window_length, target_indicator, escalations_only=True, date_column='date', location_column='adm2_pcode', rules_reversed=False, weight=0.5, selected_b=False, plot=False, set_boundaries=False, lower_boundary=0, upper_boundary=1, keep_all=True, use_crop_calendar=False):
    # make a copy of the self.dataframe
    df = df.copy()
    # run the given method
    if isinstance(window_length, list):
        df = method(df, f, window_length=window_length[0], window_length2=window_length[1], date_column=date_column, location_column=location_column, use_crop_calendar=use_crop_calendar)
    else:
        df = method(df, f, window_length=window_length, date_column=date_column, location_column=location_column, use_crop_calendar=use_crop_calendar)
    # choose whether or not to focus on escalations only
    if escalations_only:
        df = get_escalations(df, target_indicator, keep_all=keep_all)
    # filter out the rows where our target_indicator is -1 (missing rows, no data available)
    df = df[df[target_indicator] >= 0]
    # define y_real and y_score
    y_real = df[target_indicator]
    y_score = df[f'_indicator_{f}']
    # determine the lower_boundary and the upper_boundary
    lower_boundary, upper_boundary = determine_lower_and_upper_boundaries(method, rules_reversed=False, set_boundaries=set_boundaries, lower_boundary=lower_boundary, upper_boundary=upper_boundary)
    # determine the loss for each potential threshold between lower_boundary and upper_boundary and plot the results (if plot=True)
    all_scores = finding_optimal_thresholds(y_real, y_score, lower_boundary=lower_boundary, upper_boundary=upper_boundary, rules_reversed=rules_reversed, weight=weight, plot=plot)

    if not isinstance(selected_b, bool):
        boundary = selected_b
    else:
        boundary = all_scores[weight][0][0]
    y_score = binarize_indicator(y_score, boundary=boundary, rules_reversed=rules_reversed)

    row = pd.DataFrame()
    row.insert(0, 'Risk Indicator', [f])
    row.insert(1, 'Method', [method.__name__.split(f'calculate_')[1]])
    row.insert(2, 'Window Length', [window_length])
    row.insert(3, 'Boundary', [boundary])
    row.insert(4, 'Rules reversed', int(rules_reversed))
    metrics_results = get_metrics_results(y_real, y_score, weight=weight)
    for k, v in metrics_results.items():
        row.insert(row.shape[1], k, [v])
    return row    


def finding_optimal_thresholds(y_real, y_score, lower_boundary, upper_boundary, rules_reversed=False, weight=0.5, plot=False, n_steps=500):
    '''
        Determine the loss for each of the boundaries between start_point and end_point (we take a 500 linear steps by default)
        Inputs:
            y_real: outcome column
            y: the potential risk indicator column we test against y_real, before being binarized
            start_point: start point for the potential threshold
            end_point: end point for the potential threshold
            rules_reversed: Boolean, for some indicators we want to captures decreases, like drought for instance. Putting this input to True makes sure y_score is calcualted correctly.
            plot: Boolean, put to True if you wish to see the loss functions plotted
        Returns:
            all_losses: dictionary where each key is a weight factor and each value is a list of losses for each boundary between start_point and end_point
    '''
    
    fpr_fnr = []
    all_losses = dict()
    graphs = []

    if plot:
        weights = [1/3, 1/2, 2/3]
    else:
        weights = [weight]

    for idx, w in enumerate(weights):
        losses = []
        for boundary in np.linspace(lower_boundary, upper_boundary, n_steps):
            y_score_bin = binarize_indicator(y_score, boundary, rules_reversed=rules_reversed)
            losses.append([boundary, get_loss(y_real, y_score_bin, w)])

            if plot:
                if idx == 0:
                    fpr, fnr = get_fpr_and_fnr(y_real, y_score_bin)
                    fpr_fnr.append([boundary, fpr, fnr])

        if plot:
            graphs.append(go.Scatter(x=[x[0] for x in losses], y=[x[1] for x in losses], name=f'loss-function with weigh {round(w, 2)}'))

        losses.sort(key=lambda x: x[1])
        all_losses[w] = losses

    if plot:
        graphs.append(go.Scatter(x=[x[0] for x in fpr_fnr], y=[x[1] for x in fpr_fnr], name=f'FPR'))
        graphs.append(go.Scatter(x=[x[0] for x in fpr_fnr], y=[x[2] for x in fpr_fnr], name=f'FNR'))
        plot_boundaries(graphs, rules_reversed)

    if plot:
        for key, value in all_losses.items():
            print(f'weight: {round(key, 2)} has minimum of {round(value[0][1], 2)} at {value[0][0]}')

    return all_losses


def binarize_indicator(y_score, boundary, rules_reversed=False):
    '''
        Determine y_score by splitting the current risk indicator column based on the given boundary into 1's and 0's
        Input:
            y_real: outcome column
            y_score: the potential risk indicator column we test against y_real, before being binarized
            boundary: threshold boundary candidate
            rules_reversed: Boolean, for some indicators we want to captures decreases, like drought for instance. Putting this input to True makes sure y_score is calcualted correctly.
        Returns: 
            y_score: binarized column which we will map against y_real
    '''
    if not rules_reversed:
        return np.select([(y_score >= boundary), (y_score < boundary)], [1, 0])
    else:
        return np.select([(y_score <= boundary), (y_score > boundary)], [1, 0])
    

def plot_boundaries(graphs, rules_reversed):
    layout = go.Layout(
        title='LOSS FUNCTION',
        xaxis=dict(title='THRESHOLD', titlefont=dict(family='Courier New, monospace', size=18,color='#7f7f7f')),
        yaxis=dict(title='LOSS', titlefont=dict(family='Courier New, monospace', size=18, color='#7f7f7f')),
        showlegend=True,
    )
    fig = go.Figure(data=graphs, layout=layout)
    if rules_reversed:
        fig['layout']['xaxis']['autorange'] = "reversed"
    plotly.offline.iplot(fig, filename='shapes-lines')