import numpy as np
import pandas as pd


def read_jmr_master_dataset(dataset_file_location, date_column='date', location_column='adm2_pcode'):
    '''
        Reads the jmr_master_dataset.csv into a pandas dataframe and returns this dataframe
    '''
    df = pd.read_csv(dataset_file_location, float_precision='high')

    # make sure the date_column has a date format
    df[date_column] = pd.to_datetime(df[date_column]).dt.date
    number_of_rows_before = df.shape[0]

    # order by location_column and date_column and remove duplicates (NOTE: these duplicates should not exist in the first place)
    df = df.\
        sort_values(by=[location_column, date_column], ascending=True).\
        drop_duplicates(
            subset=[location_column, date_column],
            keep='first'
            ).\
        reset_index(drop=True)

    if number_of_rows_before != df.shape[0]:
        print(f'\n    Duplicate rows found in {dataset_file_location} ({number_of_rows_before} rows before dropping duplicates, {df.shape[0]} rows after)\n')

    return df


def add_target_indicator(df, target_indicator, date_column='date', location_column='adm2_pcode', high_risk_cutoff=4):
    '''
        Set a target indicator and get target indicator columns
    '''
    if 'fews' in target_indicator.lower():
        return get_fews_indicator(df, target_indicator, date_column=date_column, location_column=location_column, high_risk_cutoff=high_risk_cutoff)
    else:
        print(f'Unknown target indicator given: {target_indicator}')


def get_fews_indicator(df, target_indicator, date_column='date', location_column='adm2_pcode', high_risk_cutoff=4):
    '''
        FEWS NET indicator
        This function creates two columns:
        _fews_before_ha:             fews_ipc_weighted adjusted for provided Humanitarian Aid
        _target_indicator_fews:     when fews_before_ha >= 4 --> 1, when _fews_before_ha <= 3 --> 0, when _fews_before_ha = NULL --> -1, input name is the column name for this indicator
    '''

    # adjust for Humanitarian Aid
    df['_fews_before_ha'] = (df['fews_ipc_weighted'] + df['fews_ha']).round(0)
    # determine target definitions
    df[target_indicator] = np.select([df['_fews_before_ha'] >= high_risk_cutoff, df['_fews_before_ha'] < high_risk_cutoff, df['_fews_before_ha'].isna()], [1, 0, -1])

    # previous period
    df = df.sort_values(by=[location_column, date_column], ascending=True)
    # previous period
    df['_fews_before_ha_previous_period'] = df.groupby(location_column)['_fews_before_ha'].ffill().shift()
    # delete all previous periods that are now filled in because of the ffill()
    df['_fews_before_ha_previous_period'] = df['_fews_before_ha_previous_period'].mask(df['_fews_before_ha'].isna())
    df['_fews_target_previous_period'] = np.select([df['_fews_before_ha_previous_period'] >= high_risk_cutoff, df['_fews_before_ha_previous_period'] < high_risk_cutoff, df['_fews_before_ha_previous_period'].isna()], [1, 0, -1])
    # fews escalation is where the previous period was a 0
    df[f'{target_indicator}_escalations'] = np.select([(df['_fews_target_previous_period']==0)], [1])
    return df


def get_escalations(df, target_indicator, keep_all=True):
    '''
        Returns the Dataframe df, where all non-escalation rows are filtered out
        Inputs:
            df: Dataframe
            target_indicator: name of the target indicator (FEWS, IPC, ..)
            keep_all: default is False, which means we trim down the number of rows where target_indicator = 0 to be at most 4 times the number of rows where target_indicator = 1
        Returns:
            dataframe with only escalations
    '''
    seed = 1
    # we are only modeling for escalations and we do not want to keep rows where the target indicator is empty
    df = df[df[f'{target_indicator}_escalations']==1]
    df = df[df[target_indicator] >= 0]

    # if the difference between  1's and 0's in the target indicator is to high, we want to take a random sample of the largest of the two sets
    # to make sure the rate is at most 1:4
    if keep_all:
        return df
    else:
        # number of non-normals
        df_non_normals = df[df[target_indicator]==1]
        df_normals = df[df[target_indicator]==0]
        nr_non_normals = df_non_normals.shape[0]
        df_normals_sample = df_normals.sample(4*nr_non_normals, random_state=seed)
        return pd.concat([df_non_normals, df_normals_sample])