from sklearn import metrics


def get_loss(y_real, y_score, weight=0.5):
    '''
        Returns the loss score, which is defined as a weighted average of the False Positive Rate (FPR) and the False Negative Rate (FNR)
        The input 'w' is the weight assigned to how heavy we count FPR, 1-w is the weight corresponding to FNR
        Input:
            y_real: outcome column
            y_score: the potential risk indicator column we test against y_real
            w: weight, value between 0 and 1. Higher numbers for w put more weight on the FPR
        Returns:
            loss, value between 0 and 1, over which we are minimizing
    '''
    TN, FP, FN, TP = metrics.confusion_matrix(y_real, y_score).ravel()
    FPR = FP/(FP+TN)
    FNR = FN/(FN+TP)
    loss = weight * FNR + (1-weight) * FPR

    return loss


def get_brier_and_r_squared(y_real, y_score):
    # Brier score (mean squared error)
    err = y_score - y_real
    err = err**2
    BS = 1/err.shape[0] * err.sum()
    # r2
    log_loss = metrics.log_loss(y_real, y_score)
    log_loss_naive = metrics.log_loss(y_real, [0.5] * y_real.shape[0])
    r2 = 1 - log_loss/log_loss_naive
    return BS, r2


def get_metrics_results(y_real, y_score, weight=0.5):
    TN, FP, FN, TP = metrics.confusion_matrix(y_real, y_score).ravel()
    return {
        'True Positives': TP,
        'False Positives': FP,
        'True Negatives': TN,
        'False Negatives': FN, 
        'Accuracy': metrics.accuracy_score(y_real, y_score),
        'Balanced Accuracy': metrics.balanced_accuracy_score(y_real, y_score),
        'Precision': metrics.precision_score(y_real, y_score),
        'F1': metrics.f1_score(y_real, y_score),
        'Kappa': metrics.cohen_kappa_score(y_real, y_score),
        'FPR': FP/(FP+TN),
        'FNR': FN/(FN+TP),
        'Loss': get_loss(y_real, y_score, weight=0.5),
        'Loss (w=1/3)': get_loss(y_real, y_score, weight=1/3),
        'Loss (w=2/3)': get_loss(y_real, y_score, weight=2/3),
        }


def get_fpr_and_fnr(y_real, y_score_bin):
    TN, FP, FN, TP = metrics.confusion_matrix(y_real, y_score_bin).ravel()
    fpr, fnr = FP/(FP+TN), FN/(FN+TP)
    return fpr, fnr